from abc import ABC, abstractmethod
from typing import List, Optional

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.aggregate import IAggregate
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.aggregate import EventStream


class IEventRepository(ABC):
    @abstractmethod
    async def save_async(self, aggregate: IAggregate):
        pass

    @abstractmethod
    async def get_by_aggregate(self, aggregate_id: IAggregate) -> EventStream:
        pass

    @abstractmethod
    async def get_by_aggregates(self, aggregate_ids: List[IAggregate]) -> Optional[List[EventStream]]:
        pass



from typing import Generic, TypeVar

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate


TAggregate = TypeVar('TAggregate', bound=IAggregate)

class IAggregateDTO(Generic[TAggregate]):
    @property
    def PartitionKey(self) -> str:
        raise NotImplementedError

    def GetDTOFromEntity(self, pEntity: TAggregate):
        raise NotImplementedError

    def MappingAggregate(self):
        raise NotImplementedError

    def ToAggregate(self) -> TAggregate:
        raise NotImplementedError



from dataclasses import dataclass
from typing import List, Optional
from uuid import UUID
import json
from abc import ABC, abstractmethod

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate
from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent
from src.core.domain.example_agg.example_aggregate import ExampleAggregate

@dataclass
class RolesUpdated:
    previous_roles: List[UUID]
    new_roles: List[UUID]

class ExampleRolesChanged(DomainEvent):
    def __init__(self, user=None, previous_ids=None, user_request=None):
        super().__init__(user_request)
        if user and previous_ids:
            user_registered = RolesUpdated(previous_ids, user.roles)
            self.j_data = json.dumps(user_registered, default=lambda o: str(o) if isinstance(o, UUID) else o.__dict__)

    async def handle(self, aggregate: IAggregate):
        if isinstance(self.source_aggregate_root, ExampleAggregate):
            example_agg = self.source_aggregate_root
            # Run some code with the handle



from dataclasses import dataclass
from enum import Enum


@dataclass(frozen=True)
class ExampleAssociatedCompany:
    type: str
    scac_code: str


class TypeAssociated(Enum):
    YARD_OWNER = "YardOwner"
    CARRIER = "Carrier"


@dataclass(frozen=True)
class AssociatedCarrierCompany(ExampleAssociatedCompany):
    scac_code: str

    def __init__(self, scac_code: str):
        super().__init__(TypeAssociated.CARRIER.value, scac_code)

@dataclass(frozen=True)
class AssociatedYardOwnerCompany(ExampleAssociatedCompany):
    scac_code: str

    def __init__(self, scac_code: str):
        super().__init__(TypeAssociated.YARD_OWNER.value, scac_code)



import logging
from typing import Any, List, Optional, TypeVar, Generic
from fastapi import HTTPException
from azure.cosmos import CosmosClient, PartitionKey, exceptions as cosmos_exceptions
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate
from faimly_t_ddd_core.common.faimly_t_python_framework.repository.icollection_repository import ICollectionRepository
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.cosmos_db_config import CosmosDbConfig
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.iaggregate_dto import IAggregateDTO

TAggregate = TypeVar('TAggregate', bound=IAggregate)
TAggregateDTO = TypeVar('TAggregateDTO', bound=IAggregateDTO)

# class ExampleRepository(CosmosCollectionRepository):
#     def __init__():
#         super.__init__("ExampleDBCosmos", "EventsExampleDb")

class CosmosCollectionRepository(ICollectionRepository[TAggregate, TAggregateDTO]):
    def __init__(self, config: CosmosDbConfig, event_repository: Any):
        self.config = config
        self.event_repository = event_repository

        self.client = CosmosClient.from_connection_string(
            config.connection_string,
            consistency_level="Session"
        )

        self.database = self.client.get_database_client(config.database_name)
        self.container = self.database.get_container_client(config.container)

    async def save_async(self, aggregate: TAggregate):
        try:
            await self.event_repository.save_async(aggregate)

            entity_dto = TAggregateDTO()
            entity_dto.GetDTOFromEntity(aggregate)

            await self.container.upsert_item(
                item=entity_dto.dict(by_alias=True),
                partition_key=PartitionKey(entity_dto.PartitionKey)
            )
        except cosmos_exceptions.CosmosHttpResponseError as ex:
            logging.error(f"CosmosException saving the aggregate {aggregate} with {aggregate.Id}, "
                         f"and StatusCode {ex.status_code} - SubStatusCode {ex.sub_status_code}, "
                         f"and message {ex.message}")
            raise HTTPException(status_code=500, detail=f"Error saving the aggregate: {ex.message}")
        except Exception as ex:
            raise HTTPException(status_code=500, detail=str(ex))

    async def get_by_specification_async(self, specification: str, include_events: bool = True) -> Optional[List[TAggregate]]:
        query = f"SELECT * FROM c WHERE {specification}"
        items = self.container.query_items(query=query, enable_cross_partition_query=True)

        results = []
        async for item in items:
            dto = TAggregateDTO(**item)
            aggregate = dto.ToAggregate()
            results.append(aggregate)

        if include_events:
            aggregates = [result for result in results if result is not None]
            event_streams = await self.event_repository.get_by_aggregate(aggregates)
            for result in results:
                event_stream = next((x for x in event_streams if x.IdAggregate == result.Id), None)
                if event_stream:
                    result.attach_event_stream(event_stream)

        return results

    async def delete_async(self, aggregate: TAggregate):
        try:
            await self.container.delete_item(
                item=aggregate.Id,
                partition_key=PartitionKey(aggregate.PartitionKey)
            )
        except cosmos_exceptions.CosmosHttpResponseError as ex:
            logging.error(f"CosmosException deleting the aggregate {aggregate} with {aggregate.Id}, "
                         f"and StatusCode {ex.status_code} - SubStatusCode {ex.sub_status_code}, "
                         f"and message {ex.message}")
            raise HTTPException(status_code=500, detail=f"Error deleting the aggregate: {ex.message}")
        except Exception as ex:
            raise HTTPException(status_code=500, detail=str(ex))



import pytest
import uuid
from typing import Optional

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.aggregate import Aggregate, EventStream
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.entity import IEntity
from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent, IAggregate

# Definimos una implementación básica de IEntity para pruebas
class MockEntityRoot(IEntity):
    def __init__(self, id: uuid.UUID):
        self._id = id

    @property
    def id(self) -> uuid.UUID:
        return self._id

    @id.setter
    def id(self, value: uuid.UUID):
        self._id = value

# Definimos un evento de dominio ficticio para pruebas
class MockDomainEvent(DomainEvent):
    async def handle(self, aggregate: IAggregate):
        # Implementar la lógica específica para el manejo del evento en pruebas
        pass

# Definimos una implementación de Aggregate para pruebas
class MockAggregate(Aggregate[MockEntityRoot]):
    def __init__(self, root: MockEntityRoot, id: Optional[uuid.UUID] = None):
        super().__init__(root, id)

@pytest.fixture
def setup_aggregate():
    root_entity = MockEntityRoot(uuid.uuid4())
    aggregate = MockAggregate(root_entity, root_entity.id)
    return aggregate

def test_aggregate_initialization(setup_aggregate):
    aggregate = setup_aggregate
    assert isinstance(aggregate, MockAggregate)
    assert aggregate.root is not None
    assert isinstance(aggregate.aggregate_event_stream, EventStream)
    assert aggregate.id == aggregate.aggregate_event_stream.id_aggregate

def test_add_event(setup_aggregate):
    aggregate = setup_aggregate
    event = MockDomainEvent(user='test_user')
    aggregate.aggregate_event_stream.add_event(event)
    assert len(aggregate.aggregate_event_stream.pending_events) == 1

def test_pull_domain_events(setup_aggregate):
    aggregate = setup_aggregate
    event = MockDomainEvent(user='test_user')
    aggregate.aggregate_event_stream.add_event(event)
    pulled_events = aggregate.pull_domain_events()
    assert len(pulled_events) == 1
    assert isinstance(pulled_events[0], DomainEvent)

def test_attach_event_stream(setup_aggregate):
    aggregate = setup_aggregate
    new_event_stream = EventStream(aggregate.id)
    aggregate.attach_event_stream(new_event_stream)
    assert aggregate.aggregate_event_stream == new_event_stream

@pytest.mark.asyncio
async def test_process_event_async(setup_aggregate):
    aggregate = setup_aggregate
    event = MockDomainEvent(user='test_user')
    aggregate.process_event_async(event)
    assert len(aggregate.aggregate_event_stream.pending_events) == 1



from typing import Optional, List
from uuid import UUID
import uuid

from src.core.domain.example_agg.entities.example import Example

from src.core.domain.example_agg.value_objects.example_associated_company import ExampleAssociatedCompany
from src.core.domain.example_agg.value_objects.example_contact_information import ExampleContactInformation
from src.core.domain.example_agg.value_objects.states.example_created_state import ExampleCreatedState
from src.core.domain.example_agg.value_objects.states.example_i_user_state import ExampleIState

class ExampleBuilder:
    def __init__(self):
        self._user: Optional[Example] = None
        self._user_state: Optional[ExampleIState] = None
        self._associated_company: Optional[ExampleAssociatedCompany] = None
        self._roles: Optional[List[UUID]] = None

    def with_user(self, name: str, email: str, phone_number: Optional[int] = None, id: Optional[UUID] = None) -> 'ExampleBuilder':
        if id is None:
            id = uuid.uuid4()

        self._user = Example(id)
        self._user.contact_information = ExampleContactInformation(name, email, phone_number)
        return self

    def from_dto(self, user: Example, role_ids: Optional[List[UUID]], id_user: UUID) -> 'ExampleBuilder':
        self._user = Example(id_user)
        self._user.contact_information = user.contact_information
        self._user.url_avatar_img = user.url_avatar_img
        self._roles = role_ids
        return self

    def with_associated_entity(self, associated_company: ExampleAssociatedCompany) -> 'ExampleBuilder':
        self._associated_company = associated_company
        return self

    def with_state(self, user_state: ExampleIState) -> 'ExampleBuilder':
        self._user_state = user_state
        return self

    def build(self):
        from src.core.domain.example_agg.example_aggregate import ExampleAggregate
        
        if self._user is None:
            raise ValueError("User must be specified.")

        if self._associated_company is None:
            raise ValueError("User must have an associated Company (Carrier or YardOwner)")

        if self._user_state is None:
            self._user_state = ExampleCreatedState(self._user)

        return ExampleAggregate(
            self._user,
            self._associated_company,
            self._roles,
            self._user_state
        )



from pydantic import BaseModel

class EndPointsPaymentBillingOptions(BaseModel):
    http_endpoint: str
    api_header: str
    api_key: str

    class Config:
        fields = {
            'http_endpoint': 'HttpEndpoint',
            'api_header': 'ApiHeader',
            'api_key': 'ApiKey',
        }


import enum
from typing import List, Optional

from pydantic import BaseModel

from src.core.domain.example_agg.value_objects.example_associated_company import ExampleAssociatedCompany

# Define los estados asociados
class TypeAssociated(enum.Enum):
    Carrier = "Carrier"
    YardOwner = "YardOwner"

# Define los estados del usuario
class UserState(enum.Enum):
    Created = "Created"
    Approved = "Approved"

# Define las clases de ejemplo para los agregados y DTOs
class AssociatedCompanyDTO(BaseModel):
    Type: str
    ScacCode: str

    def __init__(self, pCompanyToTransform: Optional['ExampleAssociatedCompany'] = None, **data):
        super().__init__(**data)
        if pCompanyToTransform:
            self.Type = pCompanyToTransform.Type
            self.ScacCode = pCompanyToTransform.ScacCode

class ExampleAggDTO(BaseModel):
    User: 'ExampleUser'
    AssociatedCompanyId: AssociatedCompanyDTO
    RoleIds: List[str]  # Usando str en lugar de Guid
    StatusUser: str

    @property
    def ScacOwner(self) -> str:
        return self.AssociatedCompanyId.ScacCode

    @property
    def PartitionKey(self) -> str:
        return self.ScacOwner

    def MappingAggregate(self, aggregate: 'UserExampleAggregate'):
        self.User = aggregate.User
        self.StatusUser = aggregate.State.StateEnum.name
        self.RoleIds = aggregate.Roles
        self.AssociatedCompanyId = AssociatedCompanyDTO(pCompanyToTransform=aggregate.AssociatedCompanyId)

    def ToAggregate(self) -> 'UserExampleAggregate':
        builder = UserExampleAggregate.Builder.FromDTO(self.User, self.RoleIds, self.id)

        associated_company_type = TypeAssociated[self.AssociatedCompanyId.Type]

        if associated_company_type == TypeAssociated.Carrier:
            builder.WithAssociatedEntity(AssociatedCarrierCompany(self.AssociatedCompanyId.ScacCode))
        elif associated_company_type == TypeAssociated.YardOwner:
            builder.WithAssociatedEntity(AssociatedYardOwnerCompany(self.AssociatedCompanyId.ScacCode))
        else:
            raise ValueError(f"The user: {self.User.ContactInformation.Name}, from the database has an invalid Associated Company Type")

        if self.StatusUser == "Created":
            builder.WithState(ExampleCreatedState())
        elif self.StatusUser == "Approved":
            builder.WithState(ExampleAuthorizeState())
        
        return builder.Build()

# Define las clases asociadas a los estados y entidades (debe ser implementado según el caso)
class ExampleUser(BaseModel):
    pass

class UserExampleAggregate(BaseModel):
    class Builder:
        @staticmethod
        def FromDTO(user, role_ids, id):
            # Implementar método de construcción
            pass
        
        def WithAssociatedEntity(self, entity):
            # Implementar método para asociar entidad
            pass

        def WithState(self, state):
            # Implementar método para establecer estado
            pass
        
        def Build(self):
            # Implementar método para construir el agregado
            pass

class AssociatedCarrierCompany(BaseModel):
    pass

class AssociatedYardOwnerCompany(BaseModel):
    pass

class ExampleCreatedState(BaseModel):
    pass

class ExampleAuthorizeState(BaseModel):
    pass



from typing import Callable, TypeVar, Generic, List
from pydantic import BaseModel, ValidationError
from abc import ABC, abstractmethod

TRequest = TypeVar('TRequest', bound=BaseModel)
TResponse = TypeVar('TResponse')


class IPipelineBehavior(ABC, Generic[TRequest, TResponse]):
    @abstractmethod
    async def handle(self, request: TRequest, next: Callable[[], TResponse]):
        pass


class ValidationBehavior(IPipelineBehavior[TRequest, TResponse]):
    def __init__(self, validators: List[Callable[[TRequest], None]]):
        self._validators = validators

    async def handle(self, request: TRequest, next: Callable[[], TResponse]):
        if self._validators:
            errors = []
            for validator in self._validators:
                try:
                    validator(request)
                except ValidationError as e:
                    errors.extend(e.errors())

            if errors:
                if isinstance(next(), Result):
                    return Result(errors=errors, valid=False)
                else:
                    raise ValidationError(errors)

        return await next()


class Result(Generic[TResponse]):
    def __init__(self, data: TResponse = None, errors: List[str] = None, valid: bool = True):
        self.data = data
        self.errors = errors or []
        self.valid = valid

    @classmethod
    def invalid(cls, errors: List[str]):
        return cls(errors=errors, valid=False)



from dataclasses import dataclass
import json
from typing import List, Optional
from uuid import UUID
from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent
from src.core.domain.example_agg.example_aggregate import ExampleAggregate

def custom_serializer(obj):
    if isinstance(obj, UUID):
        return str(obj)  # Convert UUID to string
    elif hasattr(obj, "__dict__"):
        return obj.__dict__  # Convert object with __dict__ attribute to dictionary
    raise TypeError(f"Type {type(obj)} not serializable")

class ExampleApprovedEvent(DomainEvent):
    def __init__(self, user: str):
        self.j_data = None
        super().__init__(user)

    def handle(self, aggregate: ExampleAggregate):
        if isinstance(aggregate, ExampleAggregate):
            user_registered = UserIsApproved(
                aggregate.id,
                aggregate.user.contact_information.email,
                aggregate.roles
            )

            self.j_data = json.dumps(user_registered, default=custom_serializer)

@dataclass
class UserIsApproved:
    user_id_b2c: str
    email: str
    roles_assigned: Optional[List[str]] = None


from abc import ABC, abstractmethod
from uuid import UUID
from typing import TypeVar, Generic
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.entity import Entity
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.aggregate import Aggregate

T = TypeVar('T', bound=Entity)
U = TypeVar('U')
V = TypeVar('V')

class Aggregate(ABC):
    def __init__(self, root: Entity, id: UUID):
        self.root = root
        self.id = id

class AggregateWithState(Aggregate, Generic[T, U, V]):
    def __init__(self, root: T, id: UUID, state: U):
        super().__init__(root, id)
        self.state = state

    def change_state_to(self, new_state: V):
        new_state.handle(self)
        self.state = new_state



class BaseState(ABC):
    def __init__(self, state_enum):
        self.state_enum = state_enum

    @abstractmethod
    def handle(self, aggregate):
        pass

    def __str__(self):
        return self.state_enum.name.replace("State", "")


class IBaseState(ABC):
    @property
    @abstractmethod
    def state_enum(self):
        pass

    @abstractmethod
    def handle(self, aggregate):
        pass



from dataclasses import dataclass
from typing import List, Optional

@dataclass
class ExampleAccessRights:
    resource: str
    has_access: bool
    full_access: bool
    permissions: Optional[List[str]] = None



from abc import ABC, abstractmethod
from typing import TypeVar, Generic
from pydantic import ValidationError, BaseModel

T = TypeVar('T', bound=BaseModel)


class ISpecificationValidation(ABC, Generic[T]):
    @abstractmethod
    async def validate(self, command_to_validate: T) -> ValidationError:
        pass



from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.event_repository.event_repository import EventRepository

class ExampleEventsRepository(EventRepository):
    def __init__(self, 
                 config, 
                 logger):
        super().__init__(config, logger)



from abc import ABC
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.state_pattern.ibase_state import IBaseState
from src.core.domain.example_agg.value_objects.states.example_states import ExampleState

class ExampleIState(IBaseState[ExampleState], ABC):
    pass



from typing import List, Optional, Iterable
from uuid import UUID
from dataclasses import dataclass, field
import asyncio

# Importando clases desde los módulos correspondientes
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.aggregate import Aggregate
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.state_pattern.aggregate_with_state import AggregateWithState
from src.core.domain.example_agg.entities.example import Example
from src.core.domain.example_agg.value_objects.example_associated_company import ExampleAssociatedCompany
from src.core.domain.example_agg.value_objects.states.example_i_user_state import ExampleIState
from src.core.domain.example_agg.value_objects.states.example_states import ExampleState


@dataclass
class ExampleAggregate(Aggregate[Example], AggregateWithState[Example, ExampleIState, ExampleState]):

    def __init__(self,
                 root: Example,
                 associated_company_id: ExampleAssociatedCompany,
                 roles: Optional[Iterable[UUID]] = None,
                 state: Optional[ExampleIState] = None):
        super().__init__(root, root.id)  # Llama al constructor de Aggregate
        AggregateWithState.__init__(self, root, root.id, state)  # Llama al constructor de AggregateWithState
        self.associated_company_id = associated_company_id
        self.roles = list(roles) if roles else []

        if state:
            self.state = state
            state.handle(self)


    def approved_user(self, user_approving: str):
        from src.core.domain.example_agg.events.example_approved_event import ExampleApprovedEvent
        self.process_event_async(ExampleApprovedEvent(user_approving))

    @property
    def user(self) -> Example:
        return self.root

    @staticmethod
    def builder():
        from src.core.domain.example_agg.factories.example_builder import ExampleBuilder
        return ExampleBuilder()
    
    # def process_event_async(self, event: DomainEvent):
    #     event.handle(self)



from azure.cosmos import CosmosClient
import logging

from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.cosmos_db_config import CosmosDbConfig

class CosmosDbFacade:
    def __init__(self, config: CosmosDbConfig):
        try:
            self.client = CosmosClient(
                config.ConnectionString,
                consistency_level="Session"
            )
            self.database = self.client.get_database_client(config.DatabaseName)
            self.container = self.database.get_container_client(config.Container)
        except Exception as ex:
            logging.error(f"Error initializing Cosmos DB: {ex}")
            raise Exception(f"System cannot create the Database {config.DatabaseName}.")



from typing import Optional

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate
from src.core.domain.example_agg.value_objects.states.example_i_user_state import ExampleIState
from src.core.domain.example_agg.value_objects.states.example_states import ExampleState

class ExampleAuthorizeState(ExampleIState):
    def __init__(self, user_approving: Optional[str] = None):
        self._user_approving = user_approving
        self.state_enum = ExampleState.APPROVED

    async def handle(self, aggregate: IAggregate):
        # if isinstance(aggregate, UserExampleAggregate):
        #     await aggregate.process_event_async(
        #         ExampleUserApprovedEvent(self._user_approving)
        #     )
        pass



from typing import List, Optional
from uuid import UUID
import json
from dataclasses import dataclass

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate
from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent
from src.core.domain.example_agg.example_aggregate import ExampleAggregate

class ExampleCreatedEvent(DomainEvent):
    
    def __init__(self):
        self.j_data = None
        
    def handle(self, aggregate: ExampleAggregate):
        if isinstance(aggregate, ExampleAggregate):
            # Run some code with the handle
            user_registered = NewCreated(
                user_id_b2c=aggregate.id,
                email=aggregate.user.contact_information.email,
                name=aggregate.user.contact_information.name,
                roles_assigned=aggregate.roles
            )
            self.j_data = json.dumps(user_registered.to_dict())


@dataclass
class NewCreated:
    user_id_b2c: UUID
    email: str
    name: str
    roles_assigned: Optional[List[UUID]] = None

    def to_dict(self) -> dict:
        return {
            "userIdB2C": str(self.user_id_b2c),
            "email": self.email,
            "name": self.name,
            "rolesAssigned": [str(role) for role in self.roles_assigned] if self.roles_assigned else []
        }




from dependency_injector import containers, providers
from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event_publisher import DomainEventPublisher
from faimly_t_ddd_core.common.faimly_t_python_framework.validation_commands.validation_behavior import IPipelineBehavior, ValidationBehavior


class DIContainerCommon(containers.DeclarativeContainer):
    domain_event_publisher = providers.Factory(DomainEventPublisher)
    validation_behavior = providers.Factory(IPipelineBehavior, ValidationBehavior)
    
    def add_common_services(self):
        self.domain_event_publisher()
        self.validation_behavior()



from abc import ABC, abstractmethod
from typing import Callable, TypeVar, Generic

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate

TAggregate = TypeVar('TAggregate', bound='IAggregate')


class ISpecification(ABC, Generic[TAggregate]):
    @property
    @abstractmethod
    def where_criteria(self) -> Callable[[TAggregate], bool]:
        pass

    @property
    @abstractmethod
    def where_criteria_str(self) -> str:
        pass


class ISpecificationSingleOrDefault(ISpecification[TAggregate], ABC):
    pass


class ISpecificationToList(ISpecification[TAggregate], ABC):
    pass



from pydantic import BaseModel, Field
from typing import Optional, TypeVar, Generic
from datetime import datetime
import uuid

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.iaggregate_dto import IAggregateDTO


TAggregate = TypeVar('TAggregate', bound=IAggregate)

class AggregateCosmosDTO(BaseModel, IAggregateDTO[TAggregate], Generic[TAggregate]):
    _aggregate: Optional[TAggregate] = None
    Ts: int = Field(..., alias='_ts')
    idCosmos: str = Field(..., alias='id')
    _rid: str
    _self: str
    _etag: str = Field(..., alias='_etag')
    _attachments: str
    id: uuid.UUID = Field(default_factory=uuid.uuid4)

    class Config:
        allow_population_by_field_name = True

    @property
    def TimeStampUtc(self) -> datetime:
        return datetime.utcfromtimestamp(self.Ts)

    @property
    def PartitionKey(self) -> str:
        raise NotImplementedError

    def GetDTOFromEntity(self, pEntity: TAggregate):
        self._aggregate = pEntity
        self.idCosmos = str(self._aggregate.Id)
        self.MappingAggregate()

    def ToAggregate(self) -> TAggregate:
        raise NotImplementedError

    def MappingAggregate(self):
        raise NotImplementedError

    def __init__(self, **data):
        super().__init__(**data)
        if self.idCosmos:
            self.id = uuid.UUID(self.idCosmos)



from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.configurations_helper import ConfigurationsHelper
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.cosmos_collection_repository import CosmosCollectionRepository
from src.core.domain.repositories.i_example_repository import IExampleRepository
from src.infrastructure.data_access.dto.example_agg_dto import ExampleAggDTO, UserExampleAggregate
from src.infrastructure.data_access.repository.example_event_repository import ExampleEventsRepository

class ExampleRepository(
    CosmosCollectionRepository[UserExampleAggregate, ExampleAggDTO],
    IExampleRepository
):
    def __init__(
        self,
        p_event_repository: 'ExampleEventsRepository'
    ) -> None:
        cosmos_config = ConfigurationsHelper("ExampleDBCosmos").get_cosmos_configuration("ConnectionStrings__CosmosAccountConnStr", "ExampleDBCosmos")
        super().__init__(cosmos_config, p_event_repository)



from abc import ABC, abstractmethod
from typing import Type, TypeVar, List

from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent
from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.idomain_event_subscriber import DomainEventSubscriberBase

TEvent = TypeVar('TEvent', bound='DomainEvent')


class Aggregate:
    def pull_domain_events(self) -> List[DomainEvent]:
        pass

    def pull_domain_events_typed(self, event_type: Type[TEvent]) -> List[TEvent]:
        pass


class IDomainEventPublisher(ABC):
    @abstractmethod
    async def publish_async(self, *aggregates: Aggregate):
        pass

    @abstractmethod
    async def publish_typed_async(self, event_type: Type[TEvent], *aggregates: Aggregate):
        pass

    @abstractmethod
    async def subscribe(self, *subscribers: DomainEventSubscriberBase):
        pass



from abc import ABC, abstractmethod
from uuid import UUID

class IManageDomainEvents(ABC):
    
    def __init__(self):
        self._domain_events = {}
        
    @property
    def domain_events(self):
        return self._domain_events

    def pull_domain_events(self):
        events = list(self._domain_events.values())
        self._domain_events.clear()
        return events

    def pull_domain_events_by_type(self, event_type: type):
        events = [event for event in self._domain_events.values() if isinstance(event, event_type)]
        for event in events:
            self._domain_events.pop(event.id, None)
        return events

    def attach_event_stream(self, event_stream) -> None:
        # Implementar lógica para adjuntar el stream de eventos
        pass

    async def process_event_async(self, event_item) -> None:
        # Implementar lógica para procesar el evento de manera asíncrona
        pass

class IAggregate(IManageDomainEvents, ABC):
    
    @property
    @abstractmethod
    def id(self) -> UUID:
        pass

    @id.setter
    @abstractmethod
    def id(self, value: UUID) -> None:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass



from abc import ABC, abstractmethod
from typing import List, Generic, TypeVar, Optional
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate
from faimly_t_ddd_core.common.faimly_t_python_framework.repository.ispecification import ISpecificationSingleOrDefault, ISpecificationToList
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.iaggregate_dto import IAggregateDTO

TAggregate = TypeVar('TAggregate', bound='IAggregate')
TAggregateDTO = TypeVar('TAggregateDTO', bound=IAggregateDTO)

class ICollectionRepository(Generic[TAggregate, TAggregateDTO]):
    @abstractmethod
    async def save_async(self, aggregate: TAggregate):
        pass

    @abstractmethod
    async def save_all_async(self, aggregates_to_save: List[TAggregate]):
        pass

    @abstractmethod
    async def save_with_specification_async(
        self, entity_to_load: TAggregate, specification_find: ISpecificationSingleOrDefault[TAggregate]
    ) -> TAggregate:
        pass

    @abstractmethod
    async def get_by_specification_async(
        self, specification_where: ISpecificationToList[TAggregate], include_events: bool = True
    ) -> List[Optional[TAggregate]]:
        pass

    @abstractmethod
    async def get_single_by_specification_async(
        self, specification_where: ISpecificationSingleOrDefault[TAggregate], include_events: bool = True
    ) -> Optional[TAggregate]:
        pass

    @abstractmethod
    async def get_all_async(self) -> List[TAggregate]:
        pass

    @abstractmethod
    async def delete_async(self, aggregate: TAggregate) -> TAggregate:
        pass



import inspect
from typing import Dict, Type, TypeVar

from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent, GenericDomainEvent


T = TypeVar('T', bound=Type[DomainEvent])

class DomainEventTypeResolver:
    EventTypes: Dict[str, Type[DomainEvent]] = {}

    @staticmethod
    def register_domain_event_types(reference_type: Type[T]):
        # Obtiene el módulo del tipo de referencia
        module = inspect.getmodule(reference_type)
        if module is None:
            raise ValueError("No se pudo obtener el módulo del tipo de referencia")

        # Obtiene todas las clases en el módulo que heredan de DomainEvent y no son abstractas
        for name, obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(obj, DomainEvent) and obj is not DomainEvent:
                DomainEventTypeResolver.EventTypes[obj.__name__] = obj

    @staticmethod
    def get_event_type(type_name: str) -> Type[DomainEvent]:
        return DomainEventTypeResolver.EventTypes.get(type_name, GenericDomainEvent)




import httpx
import logging

from src.infrastructure.cross_cutting.facade.endpoint_billing_options import EndPointsPaymentBillingOptions
from src.infrastructure.cross_cutting.facade.i_message_dispatcher import ResponseDispatcher


class HttpDispatcher:
    def __init__(self, http_client: httpx.AsyncClient, http_endpoint: EndPointsPaymentBillingOptions, logger: logging.Logger):
        self._http_client = http_client
        self._http_endpoint = http_endpoint
        self._logger = logger

    async def dispatch_async(self, message: str) -> ResponseDispatcher:
        try:
            headers = {
                'Content-Type': 'application/json',
                self._http_endpoint.api_header: self._http_endpoint.api_key
            }

            response = await self._http_client.post(
                self._http_endpoint.http_endpoint,
                content=message,
                headers=headers
            )

            response_json = response.text

            return ResponseDispatcher(
                is_success=response.is_success,
                status_code=response.status_code,
                response_json=response_json
            )
        except Exception as ex:
            self._logger.error(ex)
            raise ex



from typing import Callable
from pydantic import BaseModel
from uuid import UUID

from src.core.domain.example_agg.example_aggregate import ExampleAggregate
from src.core.domain.example_agg.value_objects.states.example_states import ExampleState

class ExampleByIdSpecification(BaseModel):
    where_criteria: Callable[[ExampleAggregate], bool]
    where_criteria_str: str

    @classmethod
    def from_uuid(cls, user_aggregate_id: UUID):
        where_criteria = lambda agg: agg.id == user_aggregate_id
        where_criteria_str = f"c.id = '{user_aggregate_id}'"
        return cls(where_criteria=where_criteria, where_criteria_str=where_criteria_str)

    @classmethod
    def from_str(cls, user_aggregate_id: str):
        try:
            user_guid = UUID(user_aggregate_id)
        except ValueError:
            raise ValueError(f"To build the search by User Id the param should be a valid UUID: {user_aggregate_id}")

        where_criteria = lambda agg: agg.id == user_guid
        where_criteria_str = f"c.id = '{user_guid}'"
        return cls(where_criteria=where_criteria, where_criteria_str=where_criteria_str)

class ExampleApprovedByIdSpecification(ExampleByIdSpecification):
    @classmethod
    def from_uuid(cls, user_aggregate_id: UUID):
        where_criteria = lambda agg: agg.id == user_aggregate_id and agg.state and agg.state.state_enum == ExampleState.Approved
        where_criteria_str = f"c.id = '{user_aggregate_id}' and (c.StatusUser = '{ExampleState.Approved}' or c.StatusUser = '{ExampleState.Created}')"
        return cls(where_criteria=where_criteria, where_criteria_str=where_criteria_str)




from dataclasses import dataclass, field
from uuid import UUID, uuid4
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.entity import Entity
from src.core.domain.example_agg.value_objects.example_contact_information import ExampleContactInformation

@dataclass
class Example(Entity[UUID]):
    contact_information: ExampleContactInformation = None
    url_avatar_img: str = None

    def __init__(self, id: UUID = None, contact_information: ExampleContactInformation = None, url_avatar_img: str = None):
        if id is None:
            id = uuid4()
        super().__init__(id)
        self.contact_information = contact_information
        self.url_avatar_img = url_avatar_img

    @classmethod
    def create_with_auto_id(cls, contact_information: ExampleContactInformation, url_avatar_img: str):
        return cls(uuid4(), contact_information, url_avatar_img)



from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.cosmos_db_config import CosmosDbConfig

class ExampleDBCosmos(CosmosDbConfig):
    CONN_STRING = "ConnectionStrings__CosmosAccountConnStr"
    COSMOS_SECTION = "ExampleDBCosmos"



import os
from fastapi import HTTPException

from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.cosmos_db_config import CosmosDbConfig

def get_env_variables(prefix: str) -> dict:
    env_vars = {}
    # Recorre todas las variables de entorno
    for key, value in os.environ.items():
        # Verifica si la clave comienza con el prefijo especificado
        if key.startswith(prefix) or key.startswith("ConnectionStrings__Cosmos"):
            # Agrega la variable al diccionario
            env_vars[key] = value

    return env_vars

class ConfigurationsHelper:

    def __init__(self, key_section: str):
        self.config = get_env_variables(key_section)
    
    def get_cosmos_configuration(self, key_connection: str, key_section: str) -> CosmosDbConfig:
        connection_string = self.config.get(key_connection)
        database_name = self.config.get(f"{key_section}__DatabaseName")
        container = self.config.get(f"{key_section}__Container")
        partition_key_path = self.config.get(f"{key_section}__PartitionKeyPath")

        if not connection_string:
            raise HTTPException(status_code=500, detail=f"Cannot load the connection string: {key_connection}")

        if not database_name:
            raise HTTPException(status_code=500, detail=f"Cannot load the database name: {key_section}")

        if not container:
            raise HTTPException(status_code=500, detail=f"Cannot load the container: {key_section}")

        if not partition_key_path:
            raise HTTPException(status_code=500, detail=f"Cannot load the partition key: {key_section}")

        return CosmosDbConfig(
            connection_string=connection_string,
            database_name=database_name,
            container=container,
            partition_key_path=partition_key_path
        )

    def get_connection_string(self, key_connection: str) -> str:
        connection_string = self.config.get(key_connection)
        if not connection_string:
            raise HTTPException(status_code=500, detail=f"Cannot load the connection string: {key_connection}")
        return connection_string



from collections import defaultdict
from typing import List, Type, TypeVar, Generic

from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent

TEvent = TypeVar('TEvent', bound='DomainEvent')


class Aggregate:
    def pull_domain_events(self) -> List[DomainEvent]:
        pass

    def pull_domain_events_typed(self, event_type: Type[TEvent]) -> List[TEvent]:
        pass


class DomainEventSubscriberBase(Generic[TEvent]):
    async def notify_event_async(self, domain_event: TEvent):
        pass


class DomainEventPublisher:
    def __init__(self):
        self._subscribers = defaultdict(list)

    async def subscribe(self, *subscribers: DomainEventSubscriberBase):
        for subscriber in subscribers:
            event_type = type(subscriber).__orig_bases__[0].__args__[0]
            self._subscribers[event_type].append(subscriber)

    async def publish_async(self, *aggregates: Aggregate):
        for aggregate in aggregates:
            for event_to_publish in aggregate.pull_domain_events():
                event_type = type(event_to_publish)
                if event_type in self._subscribers:
                    for handler in self._subscribers[event_type]:
                        await handler.notify_event_async(event_to_publish)

    async def publish_typed_async(self, event_type: Type[TEvent], *aggregates: Aggregate):
        for aggregate in aggregates:
            for event_to_publish in aggregate.pull_domain_events_typed(event_type):
                if event_type in self._subscribers:
                    for handler in self._subscribers[event_type]:
                        await handler.notify_event_async(event_to_publish)



from typing import List, Dict, Type, Optional
import uuid
from azure.cosmos import CosmosClient, PartitionKey, exceptions as cosmos_exceptions
from pydantic import BaseModel
import logging
import json

from src.core.domain.example_agg.example_aggregate import ExampleAggregate
from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.aggregate import EventStream
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.cosmos_db_config import CosmosDbConfig
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.event_repository.domain_event_converter import DomainEventTypeResolver

class DomainEventStoreAdapterCosmos(BaseModel):
    id_cosmos: str
    aggregate_name: str
    events: Optional[List[Dict]] = []

    @property
    def partition_key(self) -> str:
        return self.aggregate_name

    def return_event_stream(self, aggregate_type: Type[ExampleAggregate]) -> EventStream:
        DomainEventTypeResolver.register_domain_event_types(aggregate_type)

        result = {}
        if self.events:
            for event in self.events:
                event_type = event.get("EventType")
                type_ = DomainEventTypeResolver.get_event_type(event_type)
                if type_ is None:
                    raise ValueError(f"Unknown event type: {event_type}")

                domain_event = json.loads(json.dumps(event), object_hook=type_)
                sequence = event["sequence"]
                result[sequence] = domain_event

        try:
            id_guid = uuid.UUID(self.id_cosmos)
        except ValueError:
            id_guid = None

        return EventStream(result, id_guid)

class EventRepository:
    def __init__(self, config: CosmosDbConfig):
        self.client = CosmosClient.from_connection_string(config.connection_string)
        self.container = self.client.get_database_client(config.database_name).get_container_client(config.container)
        
    async def get_by_aggregate(self, aggregate: ExampleAggregate) -> Optional[EventStream]:
        try:
            query = f"SELECT * FROM c WHERE c.id = '{aggregate.id}'"
            items = list(self.container.query_items(query=query, enable_cross_partition_query=True))

            if items:
                item = items[0]
                adapter = DomainEventStoreAdapterCosmos(**item)
                return adapter.return_event_stream(type(aggregate))
            
            return None

        except cosmos_exceptions.CosmosHttpResponseError as ex:
            logging.error(f"CosmosException getting the aggregate Event Stream by Aggregate Id {aggregate.id}: {ex}")
            raise ex

        except Exception as ex:
            logging.error(f"Exception getting the aggregate Event Stream by Aggregate Id {aggregate.id}: {ex}")
            raise ex

    async def get_by_aggregates(self, aggregates: List[ExampleAggregate]) -> Optional[List[EventStream]]:
        try:
            aggregate_ids = "', '".join([agg.id for agg in aggregates])
            query = f"SELECT * FROM c WHERE c.id IN ('{aggregate_ids}')"
            items = list(self.container.query_items(query=query, enable_cross_partition_query=True))

            if items:
                event_streams = []
                for item in items:
                    adapter = DomainEventStoreAdapterCosmos(**item)
                    event_streams.append(adapter.return_event_stream(type(aggregates[0])))
                return event_streams

            return None

        except cosmos_exceptions.CosmosHttpResponseError as ex:
            logging.error(f"CosmosException getting the aggregate Event Stream by Aggregate Ids {aggregate_ids}: {ex}")
            raise ex

        except Exception as ex:
            logging.error(f"Exception getting the aggregate Event Stream by Aggregate Ids {aggregate_ids}: {ex}")
            raise ex

    async def save_async(self, aggregate: ExampleAggregate):
        try:
            entity = DomainEventStoreAdapterCosmos(
                id_cosmos=str(aggregate.id),
                aggregate_name=type(aggregate).__name__,
                events=[{
                    "sequence": index,
                    "User": event.user if hasattr(event, "user") else None,
                    "TimeStamp": event.time_stamp.isoformat() if hasattr(event, "time_stamp") else None,
                    "Data": event.j_data,
                    "EventType": event.event_type
                } for index, event in enumerate(aggregate.domain_events)]
            )
            await self.container.upsert_item(item=entity.dict(), partition_key=PartitionKey(entity.partition_key))

        except cosmos_exceptions.CosmosHttpResponseError as ex:
            logging.error(f"CosmosException saving the aggregate Event Stream {aggregate.domain_events} with Id {aggregate.id}: {ex}")
            raise ex

        except Exception as ex:
            logging.error(f"Exception saving the aggregate Event Stream {aggregate.domain_events} with Id {aggregate.id}: {ex}")
            raise ex



import pytest
from unittest.mock import AsyncMock
from uuid import UUID

from src.infrastructure.data_access.repository.example_repository import ExampleRepository
from src.core.domain.example_agg.example_aggregate import ExampleAggregate
from src.core.domain.example_agg.specifications.example_by_id_specification import ExampleByIdSpecification
from src.core.domain.example_agg.specifications.example_by_scac_owner_specification import ExampleGetByScacOwnerSpecification
from src.core.domain.example_agg.value_objects.example_associated_company import AssociatedYardOwnerCompany
from src.core.domain.example_agg.value_objects.states.example_created_state import ExampleCreatedState

from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.configurations_helper import ConfigurationsHelper
from faimly_t_ddd_core.common.faimly_t_python_framework_data_access_cosmos.cosmos_sql.event_repository.event_repository import EventRepository

class TestUserIntegration():
    @pytest.fixture(autouse=True)
    def setup(self):        
        cosmos_event_config = ConfigurationsHelper("EventsExampleDb").get_cosmos_configuration("ConnectionStrings__CosmosAccountConnStr", "EventsExampleDb")
        event_repository = EventRepository(cosmos_event_config)
        self.example_repository = ExampleRepository(event_repository)

    @pytest.mark.asyncio
    async def test_create_new_agreement(self):
        # Arrange
        user = ExampleAggregate.builder() \
            .with_user(
                "Max Rodriguez",
                "mfrodriguez@snecorp.com",
                50377589833) \
            .with_associated_entity(
                AssociatedYardOwnerCompany("SNEA")) \
            .with_state(ExampleCreatedState()) \
            .build()

        # Act
        user.approved_user("pichi")
        await self.example_repository.save_async(user)

        # Assert
        assert user is not None

    @pytest.mark.asyncio
    async def test_get_yard_location(self):
        # Arrange
        p_guid = "0c87f4f9-e228-42e8-b5a9-aba59b6413a6"
        p_id = UUID(p_guid)
        get_user_by_id = ExampleByIdSpecification(p_id)

        # Act
        user_aggregate = await self.example_repository.get_by_specification_async(get_user_by_id)

        # Assert
        assert user_aggregate is not None
        assert len(user_aggregate.aggregate_event_stream.aggregate_events) > 0

    @pytest.mark.asyncio
    async def test_get_all_users_by_scac(self):
        # Arrange
        get_carrier_by_scac = ExampleGetByScacOwnerSpecification("SNEA")

        # Act
        carrier_found = await self.example_repository.get_by_specification_async(get_carrier_by_scac)

        # Assert
        assert carrier_found is not None
        assert len(carrier_found) > 0



from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects import IAggregate
from src.core.domain.example_agg.events.example_created_event import ExampleCreatedEvent
from src.core.domain.example_agg.value_objects.states.example_i_user_state import ExampleIState
from src.core.domain.example_agg.value_objects.states.example_states import ExampleState

class ExampleCreatedState(ExampleIState):
    @property
    def state_enum(self) -> ExampleState:
        return ExampleState.CREATED

    def handle(self, aggregate: IAggregate) -> None:
        aggregate.process_event_async(
            ExampleCreatedEvent()
        )



from abc import ABC, abstractmethod
from typing import Optional
from dataclasses import dataclass, field
from datetime import datetime

from faimly_t_ddd_core.common.faimly_t_python_framework.base_objects.IAggregate import IAggregate

# Define una clase base abstracta para eventos de dominio
@dataclass
class DomainEvent(ABC):
    user: str
    j_data: Optional[dict] = field(default=None)
    source_aggregate_root: Optional['IAggregate'] = field(default=None)
    time_stamp: datetime = field(default_factory=datetime.utcnow)
    
    @property
    def event_type(self) -> str:
        return self.__class__.__name__

    @abstractmethod
    def handle(self, aggregate: 'IAggregate'):
        pass

class GenericDomainEvent(DomainEvent):
    def __init__(self, user: str, j_data: Optional[dict] = None):
        super().__init__(user, j_data=j_data)
    
    async def handle(self, aggregate: 'IAggregate'):
        # Implementar la lógica específica para GenericDomainEvent si es necesario
        raise NotImplementedError("Handle method not implemented for GenericDomainEvent")



from dataclasses import dataclass, field
from datetime import datetime

@dataclass
class ExampleDates:
    created_at: datetime = field(default_factory=lambda: datetime.utcnow())
    updated_at: datetime = field(default_factory=lambda: datetime.utcnow())

    def __init__(self, created_at=None, updated_at=None):
        if created_at is None:
            created_at = datetime.utcnow()
        if updated_at is None:
            updated_at = datetime.utcnow()
        self.created_at = created_at
        self.updated_at = updated_at



from typing import Callable
from pydantic import BaseModel

from src.core.domain.example_agg.example_aggregate import ExampleAggregate

class ExampleGetByScacOwnerSpecification(BaseModel):
    where_criteria: Callable[[ExampleAggregate], bool]
    where_criteria_str: str

    @classmethod
    def from_scac_code(cls, p_scac_code_owner: str):
        where_criteria = lambda agg: agg.associated_company_id.scac_code == p_scac_code_owner

        if p_scac_code_owner == "SNEA":
            where_criteria_str = "c.ScacOwner IS NOT NULL"
        else:
            where_criteria_str = f"c.AssociatedCompanyId.ScacCode = '{p_scac_code_owner}'"

        return cls(where_criteria=where_criteria, where_criteria_str=where_criteria_str)




from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar('T')

class IBaseState(ABC, Generic[T]):
    @property
    @abstractmethod
    def state_enum(self) -> T:
        pass

    @abstractmethod
    def handle(self, aggregate: T) -> None:
        pass



from setuptools import setup, find_packages

setup(
    name="faimly_t_ddd_core",
    version="0.1.2",
    author="David Gonzalez",
    author_email="degonzalez2194@gmial.com",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/Faimly-T/Python-DDD-TemplateCode.git",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.11',
)



from abc import ABC, abstractmethod
from typing import Dict, Generic, List, Type, TypeVar, Optional
from dataclasses import dataclass, field
import uuid

from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent

# Definimos un tipo genérico para eventos y entidades
TEvent = TypeVar('TEvent', bound='DomainEvent')
TEntityRoot = TypeVar('TEntityRoot', bound='IEntity')

@dataclass
class EventStream:
    pending_events: Dict[int, DomainEvent] = field(default_factory=dict)
    aggregate_events: Dict[int, DomainEvent] = field(default_factory=dict)
    id_aggregate: uuid.UUID = field(default_factory=uuid.uuid4)

    @property
    def current_version(self) -> int:
        return len(self.aggregate_events) + len(self.pending_events)

    def pull_pending_events(self) -> List[DomainEvent]:
        ordered_events = sorted(self.pending_events.items(), key=lambda e: e[0])
        pending_events_list = [e[1] for e in ordered_events]

        for key, event in ordered_events:
            self.aggregate_events[key] = event

        self.pending_events.clear()

        return pending_events_list

    def add_event(self, event: DomainEvent):
        self.pending_events[self.current_version + 1] = event

    def pull_domain_events(self, event_type: Type[TEvent]) -> List[TEvent]:
        selected_pending = {
            key: event for key, event in self.pending_events.items()
            if isinstance(event, event_type)
        }

        for key, event in selected_pending.items():
            self.aggregate_events[key] = event
            del self.pending_events[key]

        return list(selected_pending.values())

    def get_domain_events(self, event_type: Type[TEvent]) -> List[TEvent]:
        selected_pending = {
            key: event for key, event in self.pending_events.items()
            if isinstance(event, event_type)
        }

        aggregate_events = {
            key: event for key, event in self.aggregate_events.items()
            if isinstance(event, event_type)
        }

        domain_events = list(selected_pending.values()) + list(aggregate_events.values())
        return [event for event in domain_events if isinstance(event, event_type)]

class IEntity(ABC):
    @property
    @abstractmethod
    def id(self) -> Optional[uuid.UUID]:
        pass

    @id.setter
    @abstractmethod
    def id(self, value: Optional[uuid.UUID]):
        pass

class Aggregate(ABC, Generic[TEntityRoot]):
    root: TEntityRoot
    aggregate_event_stream: EventStream
    id: uuid.UUID

    def __init__(self, root: TEntityRoot, id: Optional[uuid.UUID] = None):
        if root is None:
            raise ValueError("root cannot be None")
        self.root = root
        self.id = id or uuid.uuid4()
        self.aggregate_event_stream = EventStream(id_aggregate=self.id)

    def attach_event_stream(self, event_stream: EventStream):
        if len(self.aggregate_event_stream.aggregate_events) > 0:
            raise ValueError("You can't assign an EventStream and replace an existing EventStream")
        self.aggregate_event_stream = event_stream

    def process_event_async(self, event_item: DomainEvent):
        try:
            event_item.handle(self)
            self.aggregate_event_stream.add_event(event_item)
        except Exception as ex:
            # Optionally, handle the exception or log it
            raise ex

    def pull_domain_events(self) -> List[DomainEvent]:
        if len(self.aggregate_event_stream.pending_events) > 0:
            return self.aggregate_event_stream.pull_pending_events()
        return []

    def pull_domain_events_by_type(self, event_type: Type[TEvent]) -> List[TEvent]:
        return self.aggregate_event_stream.pull_domain_events(event_type)

    def get_domain_events(self, event_type: Type[TEvent]) -> List[TEvent]:
        return self.aggregate_event_stream.get_domain_events(event_type)

    def __str__(self) -> str:
        return str(self.id)
    
    @property
    def domain_events(self) -> List[DomainEvent]:
        # Combina eventos pendientes y agregados
        return list(self.aggregate_event_stream.pending_events.values()) + \
               list(self.aggregate_event_stream.aggregate_events.values())



import pytest

from src.core.domain.example_agg.example_aggregate import ExampleAggregate
from src.core.domain.example_agg.value_objects.example_associated_company import AssociatedCarrierCompany, AssociatedYardOwnerCompany
from src.core.domain.example_agg.value_objects.states.example_created_state import ExampleCreatedState
from src.core.domain.example_agg.value_objects.states.example_states import ExampleState


@pytest.fixture
def setup_builder():
    def _setup_builder(name, email, phone_number, associated_company, user_state):
        builder = ExampleAggregate.builder() \
            .with_user(name, email, phone_number) \
            .with_associated_entity(associated_company) \
            .with_state(user_state)
        return builder
    return _setup_builder

def test_build_should_create_user_aggregate_with_valid_data(setup_builder):
    # Arrange
    builder = setup_builder(
        "Max Rodriguez",
        "mfrodriguez@snecorp.com",
        50377589833,
        AssociatedYardOwnerCompany("SNEA"),
        ExampleCreatedState()
    )
    
    # Act
    user_aggregate = builder.build()

    # Assert
    assert user_aggregate is not None
    assert user_aggregate.user.contact_information.name == "Max Rodriguez"
    assert user_aggregate.user.contact_information.email == "mfrodriguez@snecorp.com"
    assert isinstance(user_aggregate.associated_company_id, AssociatedYardOwnerCompany)
    assert user_aggregate.user.contact_information.phone_number == 50377589833
    assert user_aggregate.state.state_enum == ExampleState.CREATED

@pytest.mark.asyncio
async def test_process_event_enable_user(setup_builder):
    # Arrange
    builder = setup_builder(
        "Max Rodriguez",
        "mfrodriguez@snecorp.com",
        50377589833,
        AssociatedYardOwnerCompany("SNEA"),
        ExampleCreatedState()
    )
    
    # Act
    user_aggregate = builder.build()
    user_aggregate.approved_user("pichi")

    # Assert
    assert user_aggregate is not None
    assert user_aggregate.user.contact_information.name == "Max Rodriguez"
    assert user_aggregate.user.contact_information.email == "mfrodriguez@snecorp.com"
    assert isinstance(user_aggregate.associated_company_id, AssociatedYardOwnerCompany)
    assert user_aggregate.user.contact_information.phone_number == 50377589833
    assert user_aggregate.state.state_enum == ExampleState.CREATED

def test_build_should_throw_exception_when_user_not_specified():
    # Arrange
    builder = ExampleAggregate.builder()

    # Act & Assert
    with pytest.raises(ValueError, match="User must be specified."):
        builder.build()

def test_user_should_have_contact_information(setup_builder):
    # Arrange
    builder = setup_builder(
        "Alice Brown",
        "alicebrown@example.com",
        1122334455,
        AssociatedYardOwnerCompany("SNEA"),
        ExampleCreatedState()
    )
    
    # Act
    user_aggregate = builder.build()

    # Assert
    assert user_aggregate.user.contact_information is not None
    assert user_aggregate.user.contact_information.name == "Alice Brown"
    assert user_aggregate.user.contact_information.email == "alicebrown@example.com"
    assert user_aggregate.user.contact_information.phone_number == 1122334455

def test_user_should_have_default_status(setup_builder):
    # Arrange
    builder = setup_builder(
        "Bob White",
        "bobwhite@example.com",
        2233445566,
        AssociatedCarrierCompany("SNEA"),
        ExampleCreatedState()
    )
    
    # Act
    user_aggregate = builder.build()

    # Assert
    assert user_aggregate.state is not None
    assert user_aggregate.state.state_enum == ExampleState.CREATED



from abc import ABC, abstractmethod

from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent

class IDomainEventSubscriber(ABC):
    @abstractmethod
    async def notify_event_async(self, domain_event: DomainEvent):
        pass


class IDomainEventSubscriberAsync(IDomainEventSubscriber, ABC):
    @abstractmethod
    async def notify_event_async(self, domain_event: DomainEvent):
        pass


class DomainEventSubscriberBase(IDomainEventSubscriberAsync):
    @abstractmethod
    async def notify_event_async(self, domain_event: DomainEvent):
        pass

    async def notify_event_async(self, domain_event: DomainEvent):
        if isinstance(domain_event, self.__orig_bases__[0].__args__[0]):
            await self.notify_event_async(domain_event)
        else:
            raise ValueError(
                f"Invalid event type: {domain_event.__class__.__name__}, "
                f"when I'm trying to handle by {self.__class__.__name__}"
            )



from abc import ABC, abstractmethod


class BaseState(ABC):
    def __init__(self, state_enum):
        self.state_enum = state_enum

    @abstractmethod
    def handle(self, aggregate):
        pass

    def __str__(self):
        return self.state_enum.name.replace("State", "")



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Optional
from dataclasses import dataclass, field
import json

# Definimos un tipo genérico para el identificador
TId = TypeVar('TId')

@dataclass
class IEntity(ABC):
    @property
    @abstractmethod
    def id(self) -> Optional[object]:
        """
        The identifier of the entity.
        """
        pass

    @id.setter
    @abstractmethod
    def id(self, value: Optional[object]):
        pass

class Entity(ABC, Generic[TId]):
    id: Optional[TId] = field(default=None, repr=False)

    def __init__(self, id: Optional[TId] = None):
        if id is None:
            raise ValueError("id cannot be None")
        self.id = id

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Entity):
            return self.id == other.id
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        if self.id is None:
            raise ValueError("Cannot hash an entity with a None id")
        return hash(self.id)
    
    def __str__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

# Uso de la biblioteca json para la serialización
def to_json(entity: Entity) -> str:
    return json.dumps(entity, default=lambda o: o.__dict__)

def from_json(data: str, cls: type) -> Entity:
    data_dict = json.loads(data)
    return cls(**data_dict)



from datetime import datetime
import pytz

class DateHelper:
    date_formats = [
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%fZ",
        "%Y-%m-%dT%H:%M:%S",
        "%d/%m/%Y %H:%M:%S"
    ]

    @staticmethod
    def parse_date_time(p_date_time: str) -> datetime:
        for date_format in DateHelper.date_formats:
            try:
                return datetime.strptime(p_date_time, date_format).replace(tzinfo=pytz.UTC)
            except ValueError:
                continue
        raise ValueError("Invalid date format.")

    @staticmethod
    def get_date_time_utc(p_date_time: datetime) -> datetime:
        if p_date_time.tzinfo is None:
            return pytz.UTC.localize(p_date_time)
        return p_date_time.astimezone(pytz.UTC)

    @staticmethod
    def convert_to_pst(p_date_time: datetime) -> datetime:
        date_time_utc = DateHelper.get_date_time_utc(p_date_time)
        pst_time_zone = pytz.timezone('US/Pacific')
        return date_time_utc.astimezone(pst_time_zone)

    @staticmethod
    def convert_to_str(p_date_time: datetime, format: str = None) -> str:
        if not format:
            format = "%Y-%m-%d %H:%M:%S"
        return p_date_time.strftime(format)



from typing import Dict, List, Type, TypeVar, Union
from dataclasses import dataclass, field
import uuid

from faimly_t_ddd_core.common.faimly_t_python_framework.domain_events.domain_event import DomainEvent

TEvent = TypeVar('TEvent', bound=DomainEvent)

@dataclass
class EventStream:
    pending_events: Dict[int, DomainEvent] = field(default_factory=dict)
    aggregate_events: Dict[int, DomainEvent] = field(default_factory=dict)
    id_aggregate: uuid.UUID = field(default_factory=uuid.uuid4)
    
    @property
    def current_version(self) -> int:
        return len(self.aggregate_events) + len(self.pending_events)

    def __init__(self, id_aggregate: uuid.UUID):
        self.id_aggregate = id_aggregate
        self.pending_events = {}
        self.aggregate_events = {}

    def __init__(self, p_events_historical: Dict[int, DomainEvent], p_id_cosmos_aggregate: uuid.UUID):
        self.id_aggregate = p_id_cosmos_aggregate
        self.pending_events = {}
        self.aggregate_events = p_events_historical

    def pull_pending_events(self) -> List[DomainEvent]:
        # Order the events by key in ascending order
        ordered_events = sorted(self.pending_events.items(), key=lambda e: e[0])
        pending_events_list = [e[1] for e in ordered_events]
        
        # Attach the events to the aggregate events
        for key, event in ordered_events:
            self.aggregate_events[key] = event
        
        self.pending_events.clear()
        
        return pending_events_list

    def add_event(self, event: DomainEvent):
        self.pending_events[self.current_version + 1] = event

    def pull_domain_events(self, event_type: Type[TEvent]) -> List[TEvent]:
        selected_pending = {
            key: event for key, event in self.pending_events.items()
            if isinstance(event, event_type)
        }
        
        for key, event in selected_pending.items():
            self.aggregate_events[key] = event
            del self.pending_events[key]
        
        return list(selected_pending.values())

    def get_domain_events(self, event_type: Type[TEvent]) -> List[TEvent]:
        selected_pending = {
            key: event for key, event in self.pending_events.items()
            if isinstance(event, event_type)
        }

        aggregate_events = {
            key: event for key, event in self.aggregate_events.items()
            if isinstance(event, event_type)
        }

        domain_events = list(selected_pending.values()) + list(aggregate_events.values())
        
        return [event for event in domain_events if isinstance(event, event_type)]
